@@ -46,8 +46,8 @@ import org.apache.http.HttpVersion;
 import org.apache.http.MethodNotSupportedException;
 import org.apache.http.ProtocolException;
 import org.apache.http.UnsupportedHttpVersionException;
-import org.apache.http.annotation.ThreadingBehavior;
 import org.apache.http.annotation.Contract;
+import org.apache.http.annotation.ThreadingBehavior;
 import org.apache.http.concurrent.Cancellable;
 import org.apache.http.entity.ContentType;
 import org.apache.http.impl.DefaultConnectionReuseStrategy;
@@ -277,10 +277,10 @@ public class HttpAsyncService implements NHttpServerEventHandler {
     @Override
     public void exception(
             final NHttpServerConnection conn, final Exception cause) {
+        log(cause);
         final State state = getState(conn);
         if (state == null) {
             shutdownConnection(conn);
-            log(cause);
             return;
         }
         state.setTerminated();
@@ -400,9 +400,14 @@ public class HttpAsyncService implements NHttpServerEventHandler {
                 "Unexpected response state %s", state.getResponseState());
 
         if (state.getRequestState() == MessageState.ACK_EXPECTED) {
-            final Outgoing outgoing = state.getOutgoing();
-            Asserts.notNull(outgoing, "Outgoing response");
-
+            final Outgoing outgoing;
+            synchronized (state) {
+                outgoing = state.getOutgoing();
+                if (outgoing == null) {
+                    conn.suspendOutput();
+                    return;
+                }
+            }
             final HttpResponse response = outgoing.getResponse();
             final int status = response.getStatusLine().getStatusCode();
             if (status == 100) {
@@ -447,9 +452,9 @@ public class HttpAsyncService implements NHttpServerEventHandler {
                     conn.suspendOutput();
                     try {
                         handler.handle(result, httpExchange, context);
-                    } catch (RuntimeException ex) {
+                    } catch (final RuntimeException ex) {
                         throw ex;
-                    } catch (Exception ex) {
+                    } catch (final Exception ex) {
                         pipeline.add(new PipelineEntry(
                             request,
                             null,
@@ -525,7 +530,8 @@ public class HttpAsyncService implements NHttpServerEventHandler {
     public void timeout(final NHttpServerConnection conn) throws IOException {
         final State state = getState(conn);
         if (state != null) {
-            closeHandlers(state, new SocketTimeoutException());
+            exception(conn, new SocketTimeoutException(
+                    String.format("%,d milliseconds timeout on connection %s", conn.getSocketTimeout(), conn)));
         }
         if (conn.getStatus() == NHttpConnection.ACTIVE) {
             conn.close();
@@ -614,6 +620,17 @@ public class HttpAsyncService implements NHttpServerEventHandler {
 
     protected HttpAsyncResponseProducer handleException(
             final Exception ex, final HttpContext context) {
+        String message = ex.getMessage();
+        if (message == null) {
+            message = ex.toString();
+        }
+        final HttpResponse response = this.responseFactory.newHttpResponse(HttpVersion.HTTP_1_1,
+                toStatusCode(ex, context), context);
+        return new ErrorResponseProducer(response,
+                new NStringEntity(message, ContentType.DEFAULT_TEXT), false);
+    }
+
+    protected int toStatusCode(final Exception ex, final HttpContext context) {
         final int code;
         if (ex instanceof MethodNotSupportedException) {
             code = HttpStatus.SC_NOT_IMPLEMENTED;
@@ -621,17 +638,12 @@ public class HttpAsyncService implements NHttpServerEventHandler {
             code = HttpStatus.SC_HTTP_VERSION_NOT_SUPPORTED;
         } else if (ex instanceof ProtocolException) {
             code = HttpStatus.SC_BAD_REQUEST;
+        } else if (ex instanceof SocketTimeoutException) {
+            code = HttpStatus.SC_GATEWAY_TIMEOUT;
         } else {
             code = HttpStatus.SC_INTERNAL_SERVER_ERROR;
         }
-        String message = ex.getMessage();
-        if (message == null) {
-            message = ex.toString();
-        }
-        final HttpResponse response = this.responseFactory.newHttpResponse(HttpVersion.HTTP_1_1,
-                code, context);
-        return new ErrorResponseProducer(response,
-                new NStringEntity(message, ContentType.DEFAULT_TEXT), false);
+        return code;
     }
 
     /**
@@ -757,7 +769,7 @@ public class HttpAsyncService implements NHttpServerEventHandler {
             handler = (HttpAsyncRequestHandler<Object>) this.handlerMapper.lookup(request);
         }
         if (handler == null) {
-            handler = new NullRequestHandler();
+            handler = NullRequestHandler.INSTANCE;
         }
         return handler;
     }
